//////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2016, QualityLogic, Inc.
// All rights reserved.
//
// 'VEN Device Handler' and 'Arduino_Ven_Poller.ino' ("this software") are licensed under the 
// BSD 3-Clause license.
//
// Redistribution and use in source and binary forms, with or without modification, 
// are permitted provided that the following conditions are met:
//
//   * Redistributions of source code must retain the above copyright notice, this 
//     list of conditions and the following disclaimer.
//     
//   * Redistributions in binary form must reproduce the above copyright notice, 
//     this list of conditions and the following disclaimer in the documentation 
//     and/or other materials provided with the distribution.
//     
//   * Neither the name of QualityLogic nor the names of its contributors may 
//     be used to endorse or promote products derived from this software without 
//     specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
// OF SUCH DAMAGE.


#include <SoftwareSerial.h>  //Make sure to use Arduino Uno as SofwareSerial will not work with zero
#include <SmartThings.h>
 
#define PIN_THING_RX    3
#define PIN_THING_TX    2
 
 
SmartThingsCallout_t messageCallout;    // call out function forward decalaration
SmartThings smartthing(PIN_THING_RX, PIN_THING_TX, messageCallout);  // constructor
 
bool State;
bool isDebugEnabled = true;
unsigned long pollInterval = 30000; //DefaultPoll interval in milliseconds


void setup()
{
  // setup default state of global variables
  if (isDebugEnabled)
  { // setup debug serial port
     Serial.begin(115200);         // setup serial with a baud rate of 9600
     Serial.println("setup..");  // print out 'setup..' on start
  }
   
  smartthing.run();
  State = HIGH;
}
 
 
void loop()
{
	//Delay for polling interval
	millisDelay(pollInterval);


  //alternate string in message to device handler
  //not sure neccessary to alternate, but SmartThings may block if no change
  if (State == HIGH) {
	  tick();
	  State = LOW;
  }
  else {
	  tock();
	  State = HIGH;
  }


}

//toggle light when VTN polling is triggered
void blinkLight()
{
	smartthing.shieldSetLED(0, 0, 1);
	millisDelay((unsigned long) 50);
	smartthing.shieldSetLED(0, 0, 0);
	}


//send message to device handler which triggers VEN to poll VTN
void tick()
{
	if (isDebugEnabled)
		Serial.println("Inside tick()");
	blinkLight();
	smartthing.send("tick:" + String(pollInterval / 1000));        // send message to cloud
}
 
//send message to device handler which triggers VEN to poll VTN
void tock()
{
	if (isDebugEnabled)
		Serial.println("Inside tock()");
	blinkLight();
	smartthing.send("tock:" + String(pollInterval/1000));       // send message to cloud
}
 
//receive message from device handler
void messageCallout(String message)
{
	if (isDebugEnabled) {
		Serial.print("Received message: '");
		Serial.print(message);
		Serial.println("' ");
	}

	if (message.indexOf("set:") != -1) {
		pollInterval = message.substring(4).toInt() *  1000;
		if (isDebugEnabled) 
			Serial.println("pollInterval set to " + String(pollInterval));
	}
}

//non-blocking delay in milliseconds
void millisDelay(unsigned long interval)
{
	unsigned long currentMillis = millis();
	unsigned long previousMillis = millis();
		
	while ((unsigned long)(currentMillis - previousMillis) <= interval) {

		// run smartthing logic 
		smartthing.run();

		// Get snapshot of time	
		currentMillis = millis();	

		//check for change in polling interval from device handler
		if (interval != pollInterval)
			break;
	}

}
