//////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2016, QualityLogic, Inc.
// All rights reserved.
//
// 'VEN Device Handler' and 'Arduino_Ven_Poller.ino' ("this software") are licensed under the 
// BSD 3-Clause license.
//
// Redistribution and use in source and binary forms, with or without modification, 
// are permitted provided that the following conditions are met:
//
//   * Redistributions of source code must retain the above copyright notice, this 
//     list of conditions and the following disclaimer.
//     
//   * Redistributions in binary form must reproduce the above copyright notice, 
//     this list of conditions and the following disclaimer in the documentation 
//     and/or other materials provided with the distribution.
//     
//   * Neither the name of QualityLogic nor the names of its contributors may 
//     be used to endorse or promote products derived from this software without 
//     specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
// OF SUCH DAMAGE.

import groovy.transform.Field;
import java.text.SimpleDateFormat

metadata {
  definition (name: "VEN Device Handler", namespace: "", author: "QualityLogic") {
        capability "Sensor"
        capability "Actuator"
        capability "Door Control"  //closed = normal, open = moderate, opening = high, closing = emergency
        
        attribute "oadrState", "string"
        attribute "toggle", "string"  //use for Set_Poll button display
        
        command "asyncOptOut"
        command "setPollInterval"
  }

    preferences {

       //venID
       input "S_venID", "text", title: "Enter venID",
              description: "venID.", defaultValue: "TH_VEN",
              required: false, displayDuringSetup: true
       
       //vtnID
       input "S_vtnID", "text", title: "Enter vtnID",
              description: "venID", defaultValue: "TH_VTN",
              required: false, displayDuringSetup: true
              
       //VTN URL
       input "S_vtnURI", "text", title: "Enter VTN URL",
              description: "venID", defaultValue: "http://127.0.0.1:8080/OpenADR2/Simple/EiEvent",
              required: false, displayDuringSetup: true
       
       //polling interval
       input "S_pollInterval", "number", title: "Set Polling Interval (seconds)",
              description: "Polling Pnterval", defaultValue: 30,
              required: false, displayDuringSetup: true            
    }
    
  tiles {
    
        valueTile("oadrState", "device.oadrState", decoration: "flat") {
        state "oadrState", label:'${currentValue}'
    }
        
        standardTile("optOut", "device.oadrState", width: 1, height: 1, canChangeIcon: false, canChangeBackground: false) {
      state "far", label: "optOut", action: "asyncOptOut", icon: "", backgroundColor: "#79b821"
            state "near", label: "optOut", action: "asyncOptOut", icon: "", backgroundColor: "#79b821"
            state "active", label: "optOut", action: "asyncOptOut", icon: "", backgroundColor: "#79b821"
            state "done", label: "optOut", icon: "", backgroundColor: "#A0A0A0", defaultState: true
    } 
        

        standardTile("PollToggle", "device.toggle", width: 1, height: 1, canChangeIcon: false, canChangeBackground: false) {
            state "enabled", label: "Set_Poll", action: "setPollInterval", icon: "", backgroundColor: "#79b821", defaultState: true
            state "press", label: "Set_Poll", icon: "", backgroundColor: "#A0A0A0"
    } 
        
    main "oadrState"
        details(["oadrState", "optOut", "PollToggle"])
  }
}


//***************************
// VEN Configuration Settings
// Note will match any marketContext
// Note that S_vtnURI, S_venID. S_vtnID, S_pollInterval are set in phone GUI
//***************************

@Field def S_normalIntervalValue = ["0", "0.0"] //must be 0,1,2,3
@Field def S_normalAttribute = "door" 
@Field def S_normalAttributeValue = "closed"

@Field def S_moderateIntervalValue = ["1", "1.0"] //must be 0,1,2,3
@Field def S_moderateAttribute = "door" 
@Field def S_moderateAttributeValue = "open"

@Field def S_highIntervalValue = ["2", "2.0"] //must be 0,1,2,3
@Field def S_highAttribute = "door" 
@Field def S_highAttributeValue = "opening"

@Field def S_specialIntervalValue = ["3", "3.0"] //must be 0,1,2,3
@Field def S_specialAttribute = "door" 
@Field def S_specialAttributeValue = "closing"


//*************************
// oadrPoll() - Get Events
// Can't do randomization due to 40 second execution limit
//*************************
def oadrPoll() {
  log.debug "******************************"
  log.debug "OADR: Sending oadrRequestEvent"


    // oadrRequestEvent Payload - Note replyLimit = 1
    def oadrRequestEvent = """\
          |<?xml version="1.0" encoding="UTF-8"?>
          |<ns3:oadrRequestEvent
          |  xmlns="http://docs.oasis-open.org/ns/energyinterop/201110/payloads"
          |  xmlns:ns2="http://docs.oasis-open.org/ns/energyinterop/201110" xmlns:ns3="http://openadr.org/oadr-2.0a/2012/07">
          |  <eiRequestEvent>
          |    <requestID/>
          |    <ns2:venID>${S_venID}</ns2:venID>
          |    <replyLimit>1</replyLimit>
          |  </eiRequestEvent>
          |</ns3:oadrRequestEvent>
    """.stripMargin()


  def params = [
    uri: "${S_vtnURI}",
    headers: [
    'contentType': 'application/xml; charset=UTF-8'
    ],
    body: "${oadrRequestEvent}"
  ]

  def success = { response ->
    
  //Confirm correct payload received
    if (response.data[0].name != "oadrDistributeEvent"){
      log.error "OADR: *** Unexpected payload: ${response.data}"
        sendEvent(name: "oadrState", value: "Error")
        return
     }else     
      log.debug "OADR: Received oadrDistributeEvent"
    
     //Confirm no error response code
     def responseCode = response.data.eiResponse?.responseCode?.toString()
     if (responseCode != "200"){
      log.error "OADR: *** oadrDistributeEvent responseCode Error: ${responseCode}"
        sendEvent(name: "oadrState", value: "Error")
        return
     }
  
    //Confirm we have event in payload
    def eventID = response.data.oadrEvent[0]?.eiEvent?.eventDescriptor?.eventID?.toString()
  if(eventID == ""){
      log.debug "OADR: *** No Events in payload"
      //Implicit Cancellation if anything active
      sendEvent(name: "oadrState", value: "No Events")  //Modify Tile Text
      if(state.oneShotEventDone != null && state.oneShotEventDone != true){
          state.oneShotEventDone = true
            changeLoadLevel(S_normalIntervalValue[0])  //Normal Operations, implicit cancellation
      }
      return
   } 
   
   
   //Gather up info about the event
   //Keep it simple with just a couple of intervals supported
   def intervals = new String[3][2]
   def modificationNumber = response.data.oadrEvent[0]?.eiEvent?.eventDescriptor?.modificationNumber?.toInteger()
   def requestID = response.data.requestID?.toString()
   def eventStatus = response.data.oadrEvent[0]?.eiEvent?.eventDescriptor?.eventStatus?.toString()
   def dtstart = response.data.oadrEvent[0]?.eiEvent?.eiActivePeriod?.properties?.dtstart?.'date-time'?.toString()
   def duration = response.data.oadrEvent[0]?.eiEvent?.eiActivePeriod?.properties?.duration?.duration?.toString()
   def vtnID = response.data.vtnID?.toString()
   def signalName = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.signalName?.toString()
   def numberIntervals = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval?.size()
   def venID_Target = response.data.oadrEvent[0]?.eiEvent?.eiTarget?.venID?.toString()
   intervals[0][0] = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval[0]?.duration?.duration?.toString()
   intervals[0][1] = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval[0]?.signalPayload?.payloadFloat?.value?.toString()
   intervals[1][0] = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval[1]?.duration?.duration?.toString()
   intervals[1][1] = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval[1]?.signalPayload?.payloadFloat?.value?.toString()
   intervals[2][0] = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval[2]?.duration?.duration?.toString()
   intervals[2][1] = response.data.oadrEvent[0]?.eiEvent?.eiEventSignals?.eiEventSignal?.intervals?.interval[2]?.signalPayload?.payloadFloat?.value?.toString()
   def oadrResponseRequired = response.data.oadrEvent[0]?.oadrResponseRequired?.toString()
   
 
   
   //Check for errors in payload 
   def respCode ="200"
   def respDesc = "OK"
   def logDesc = ""
   def eventRespCode = "200"
   
   if(response.data.oadrEvent[1]?.eiEvent?.eventDescriptor?.eventID?.toString() != ""){ //Number of events
      respDesc = "Two or more events received when replyLimit equals 1"
        logDesc = respDesc
      respCode ="451"
   }else if (numberIntervals > 3){ //Check number of intervals
      respDesc = "This VEN only supports 3 intervals, ${numberIntervals} intervals sent"
        logDesc = respDesc
      respCode ="451"
   }else if (duration2seconds(duration) != duration2seconds(intervals[0][0]) + duration2seconds(intervals[1][0]) + duration2seconds(intervals[2][0])){ //Confirm overall duration matched sum of intervals
      respDesc = "Overall duration does not match sum of interval durations"
        logDesc = respDesc
      respCode ="454"
   }else if(vtnID != S_vtnID){    //Check vtnID
      respDesc = "vtnID (${vtnID}) does not match VEN Configuration (${S_vtnID})"
        logDesc = respDesc
      respCode ="452"
   }else if(!signalName.equalsIgnoreCase("simple")){ //Check signalName
      logDesc = "signalName (${signalName}) is not 'simple'"
      eventRespCode = "453"
   }else if (state.activeEventID == eventID && state.activeModificationNumber > modificationNumber){ //Check modificationNumber
    logDesc = "modificationNumber (${modificationNumber}) is less than previous received number (${state.activeModificationNumber})"
      eventRespCode = "450"
   }else if ((!response.data.oadrEvent[0]?.eiEvent?.eiTarget?.venID?.isEmpty() || !response.data.oadrEvent[0]?.eiEvent?.eiTarget?.resourceID?.isEmpty() ||
               !response.data.oadrEvent[0]?.eiEvent?.eiTarget?.partyID?.isEmpty() || !response.data.oadrEvent[0]?.eiEvent?.eiTarget?.groupID?.isEmpty()) &&
               venID_Target != S_venID){  //Check empty target or venID target
        logDesc = "venID (${S_venID}) not found in eiTarget subelements. Only VEN targeting is supported."
      eventRespCode = "462"
   }         
   //Return createdEvent with error if error found
   if (respCode != "200" || eventRespCode != "200"){
        log.debug "OADR: *** ${logDesc}"
        createdEvent_Response(requestID, eventID, modificationNumber, "optOut", "${respCode}", "${respDesc}", "${eventRespCode}")
        sendEvent(name: "oadrState", value: "Error")
        return
   }
     
    //Check for no change
    if (state.activeEventID == eventID && state.activeModificationNumber == modificationNumber){
        log.debug "OADR: *** Unmodified ${eventStatus} event in payload"
        
        if (eventStatus == "completed" && state.oneShotEventDone != true){
            state.oneShotEventDone = true
              changeLoadLevel(S_normalIntervalValue[0])  //Normal Operations 
        }else if (eventStatus != "cancelled" && state.optState == "optIn"){  
           //*** Evaluate intervals for value change
           intervalValueControl(dtstart, duration, intervals)
           
           //Decrement fallback resume normal by polling interval
           if (state.fallbackResumeNormalOperations > 0)
              state.fallbackResumeNormalOperations = state.fallbackResumeNormalOperations - S_pollInterval.toInteger()
        }
        if (oadrResponseRequired == "always")   
          createdEvent_Response(requestID, eventID, modificationNumber, "${state.optState}", "${respCode}", "${respDesc}", "${eventRespCode}")
        
        if(state.optState == "optIn")
          sendEvent(name: "oadrState", value: "${eventStatus}") 
        return
    }

    //Check for cancelled event
    if (eventStatus == "cancelled" && state.activeEventID == eventID && state.activeModificationNumber < modificationNumber){
        state.activeModificationNumber = modificationNumber
        log.debug "OADR: *** Cancelled event in payload, resuming normal operations"
        sendEvent(name: "oadrState", value: "${eventStatus}")  //Toggle VEN display to "DONE"
        state.oneShotEventDone = true
        if(secondsFromNow(dtstart) > 0) //Event Active
          changeLoadLevel(S_normalIntervalValue[0])  //Normal Operations
        if (oadrResponseRequired == "always")
          createdEvent_Response(requestID, eventID, modificationNumber, "optIn", "${respCode}", "${respDesc}", "${eventRespCode}")
         return
     }



     //Check for modified or new event
     if (state.activeEventID == eventID && state.activeModificationNumber < modificationNumber || (state.activeEventID != eventID && eventStatus !="completed")){
         if (state.activeEventID != eventID){
          log.debug "OADR: *** New ${eventStatus} event in payload"
            state.optState = "optIn"
            state.oneShotEventDone = false
         }
         else
          log.debug "OADR: *** Modified ${eventStatus} event in payload"

                //*** Evaluate intervals for value change
        if(state.optState == "optIn"){
          intervalValueControl(dtstart, duration, intervals)
            sendEvent(name: "oadrState", value: "${eventStatus}")  
        }
       
        //Update state variable 
        state.activeEventID = eventID
        state.activeModificationNumber = modificationNumber
        state.activeRequestID = requestID
        
        //Set fallback resume normal operations
        state.fallbackResumeNormalOperations = 0
        if (duration2seconds(duration) > 0)
          state.fallbackResumeNormalOperations = secondsFromNow(dtstart) * -1 + duration2seconds(duration)
        
       //Dump event info
       log.debug "OADR: ++ requestID ${requestID}"
       log.debug "OADR: ++ eventID ${eventID}"
       log.debug "OADR: ++ modificationNumber ${modificationNumber}"
       log.debug "OADR: ++ eventStatus ${eventStatus}"
       log.debug "OADR: ++ dtstart ${dtstart}"
       log.debug "OADR: ++ overall duration ${duration}"
       log.debug "OADR: ++ interval 1 duration ${intervals[0][0]}"
       log.debug "OADR: ++ interval 1 value: ${intervals[0][1]}"
       if (intervals[1][0] != ""){
          log.debug "OADR: ++ interval 2 duration ${intervals[1][0]}"
          log.debug "OADR: ++ interval 2 value: ${intervals[1][1]}"
            if (intervals[2][0] != ""){
                log.debug "OADR: ++ interval 3 duration ${intervals[2][0]}"
                log.debug "OADR: ++ interval 3 value: ${intervals[2][1]}"
            }
       }     
       log.debug "OADR: ++ oadrResponseRequired: ${oadrResponseRequired}"
                         
       if (oadrResponseRequired == "always")
         createdEvent_Response(requestID, eventID, modificationNumber, "${state.optState}", "${respCode}", "${respDesc}", "${eventRespCode}")
       return
     }
     else
        log.error "OADR: Unknown complete or cancelled oadrDistributeEvent received, no action taken"
  }
  
  try {
      httpPost(params, success)
   }
   catch (e) {
      
      //Decrement fallback resume normal by polling interval
     if (state.fallbackResumeNormalOperations != null && state.fallbackResumeNormalOperations > 0 && state.oneShotEventDone != true){
         state.fallbackResumeNormalOperations = state.fallbackResumeNormalOperations - S_pollInterval.toInteger()
       
         //If fallback still positive, output log message
       if(state.fallbackResumeNormalOperations > 0)
            log.debug "OADR: *** ${state.fallbackResumeNormalOperations} seconds until fallback resume normal operations"
  }

     //If fallback negative or zero, resume normal operations
     if (state.fallbackResumeNormalOperations != null && state.fallbackResumeNormalOperations <= 0 && state.oneShotEventDone != true){    
          state.oneShotEventDone = true
          changeLoadLevel(S_normalIntervalValue[0])
          log.debug "OADR: *** Fallback resume normal operation"
      }

      sendEvent(name: "oadrState", value: "No_VTN")
      log.error "OADR: oadrRequestEvent Poll Failure: $e"
   }
}

//*************************
// createdEvent_Response(requestID, eventID, modificationNumber, optState, responseCode, responseDescription, eventResponseCode)
//*************************
def createdEvent_Response(requestID, eventID, modificationNumber, optState, responseCode, responseDescription, eventResponseCode) {
  log.debug "OADR: Sending oadrCreatedEvent: " + "${optState}, " + "${responseCode} (payload), " + "${eventResponseCode} (event)"

    // oadrCreatedEvent Payload
    def oadrCreatedEvent = """\
            |<?xml version="1.0" encoding="UTF-8"?>
            |<ns3:oadrCreatedEvent
            |  xmlns="http://docs.oasis-open.org/ns/energyinterop/201110"
            |  xmlns:ns2="http://docs.oasis-open.org/ns/energyinterop/201110/payloads" xmlns:ns3="http://openadr.org/oadr-2.0a/2012/07">
            |  <ns2:eiCreatedEvent>
            |    <eiResponse>
            |      <responseCode>${responseCode}</responseCode>
            |      <responseDescription>${responseDescription}</responseDescription>
            |      <ns2:requestID/>
            |    </eiResponse>
            |    <eventResponses>
            |      <eventResponse>
            |        <responseCode>${eventResponseCode}</responseCode>
            |        <ns2:requestID>${requestID}</ns2:requestID>
            |        <qualifiedEventID>
            |          <eventID>${eventID}</eventID>
            |          <modificationNumber>${modificationNumber}</modificationNumber>
            |        </qualifiedEventID>
            |        <optType>${optState}</optType>
            |      </eventResponse>
            |    </eventResponses>
            |    <venID>${S_venID}</venID>
            |  </ns2:eiCreatedEvent>
            |</ns3:oadrCreatedEvent>
    """.stripMargin()

  def params = [
    uri: "${S_vtnURI}",
    headers: [
    'contentType': 'application/xml; charset=UTF-8'
    ],
    body: "${oadrCreatedEvent}"
  ]

  def success = { response ->
      
  //Confirm correct payload received
    if (response.data[0].name != "oadrResponse"){
      log.error "OADR: Unexpected payload: ${response.data}"
        return
     }
   
    log.debug "OADR: Received oadrResponse"
 
    //Confirm no error response code
     def responseCode1 = response.data.eiResponse?.responseCode?.toString()
     if (responseCode1 != "200"){
      log.error "OADR: oadrResponse responseCode Error: ${responseCode1}"
        return
     }
  }
  
  try {
      httpPost(params, success)
   }
   catch (e) {
      log.error "oadrCreatedEvent Send Failure: $e"
      sendEvent(name: "oadrState", value: "No Response")
   }
}

//*************************
// asyncOptOut()
//*************************
def asyncOptOut(){
 log.debug "******************************"
 log.debug "OADR: *** Async OptOut Command Received from App Layer"
 state.optState = "optOut"
 changeLoadLevel(S_normalIntervalValue[0])  //Normal Operations 
 createdEvent_Response(state.activeRequestID, state.activeEventID, state.activeModificationNumber, "${state.optState}", "200", "OK", "200")
 sendEvent(name: "oadrState", value: "Opted_Out")  
}

//*************************
// intervalValueControl(dtstart, overallDuration, intervals)
//*************************
def intervalValueControl(dtstart, overallDuration, intervals){
  //log.debug "in intervalValueControl ${dtstart} ${overallDuration} ${intervals[0][1]}  ${intervals[1][1]}  ${intervals[2][1]}"
 
    def timeDelta = secondsFromNow(dtstart)
    def interval_1_Seconds = duration2seconds(intervals[0][0])
    def interval_12_Seconds = interval_1_Seconds + duration2seconds(intervals[1][0])
    def interval_123_Seconds = interval_12_Seconds + duration2seconds(intervals[2][0])
     
  //log.debug "${timeDelta}   ${interval_1_Seconds}  ${interval_12_Seconds}  ${interval_123_Seconds}"
     
    //Check first interval
    if (timeDelta >= 0 && (timeDelta <= interval_1_Seconds || duration2seconds(overallDuration) == 0)){
      changeLoadLevel(intervals[0][1])
        log.debug "OADR: *** First Interval Active. Value = ${intervals[0][1]}"
    }
    //Check second interval, if any
    else if (timeDelta > interval_1_Seconds && timeDelta <= interval_12_Seconds){
        changeLoadLevel(intervals[1][1])
        log.debug "OADR: *** Second Interval Active. Value = ${intervals[1][1]}"
    }
    //Check third interval, if any
    else  if (timeDelta > interval_12_Seconds && timeDelta <= interval_123_Seconds){
        changeLoadLevel(intervals[2][1])
        log.debug "OADR: *** Third Interval Active. Value = ${intervals[2][1]}"
    }
}

//*************************
// changeLoadLevel(intervalValue)
//*************************
def changeLoadLevel(intervalValue){
  //log.debug "in changeLoadLevel ${intervalValue}"
    
    switch (intervalValue) {
        case [S_normalIntervalValue[0], S_normalIntervalValue[1]]: 
            sendEvent(name: "${S_normalAttribute}", value: "${S_normalAttributeValue}", isStateChange: true)  //Normal Operations
            //log.debug "in Normal"
            break
        case [S_moderateIntervalValue[0], S_moderateIntervalValue[1]]: 
            sendEvent(name: "${S_moderateAttribute}", value: "${S_moderateAttributeValue}", isStateChange: true)  //Moderate Load Shed
            //log.debug "in Moderate"
            break
        case [S_highIntervalValue[0], S_highIntervalValue[1]]:
            sendEvent(name: "${S_highAttribute}", value: "${S_highAttributeValue}", isStateChange: true)  //High Load Shed
            //log.debug "in High"
            break
        case [S_specialIntervalValue[0], S_specialIntervalValue[1]]:
            sendEvent(name: "${S_specialAttribute}", value: "${S_specialAttributeValue}", isStateChange: true)  //Special Load Shed
          break
            //log.debug "in Special"
         default:
          log.debug "OADR: Error changeLoadLevel switch default"
          
    }
}

//*************************
// duration2seconds(duration)
// Converts ISO 8061 hours, minutes, seconds to seconds
// Ignores years, months, weeks, and days greater than 9
//*************************
Integer duration2seconds(duration){
  //log.debug "duration2seconds"

    if (duration == "")  //No interval defined
      return 0

    def seconds = 0
      
    //See if there are any day values
    def d_offset = duration.indexOf('D')
    if (d_offset != -1)
      seconds = duration.substring(d_offset-1, d_offset).toInteger() * 86400
     
    //Strip off time substring
    def t_offset = duration.indexOf('T')
    def duration_timeOnly = duration.substring(t_offset + 1, duration.length())

    def h_offset = duration_timeOnly.indexOf('H')
    def m_offset = duration_timeOnly.indexOf('M')
    def s_offset = duration_timeOnly.indexOf('S')
    
    //hours
    if (h_offset != -1)
      seconds += duration_timeOnly.substring(0, h_offset).toInteger() * 3600
        
    //minutes
    if (m_offset != -1 ){
      if (h_offset != -1)
          seconds += duration_timeOnly.substring(h_offset+1, m_offset).toInteger() * 60
        else
          seconds += duration_timeOnly.substring(0, m_offset).toInteger() * 60
    }
    
    //seconds
    if (s_offset != -1 ){
      if (m_offset != -1)
          seconds += duration_timeOnly.substring(m_offset+1, s_offset).toInteger()
        else if (h_offset != -1)
          seconds += duration_timeOnly.substring(h_offset+1, s_offset).toInteger()
        else
          seconds += duration_timeOnly.substring(0, s_offset).toInteger()
     }
    
    //log.debug seconds

  return seconds
}

//*************************
// secondsFromNow(dtstartDatetime)
// Returns number of seconds since start of event
// Negative value means event has not started
// Converts dtStart and current time to epoch time, returns delta
//*************************
Integer secondsFromNow(dtstartDatetime){
    //log.debug "in secondsFromNow"
    
  SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'")
    Date date = df.parse(dtstartDatetime.substring(0, 19) + "Z")
    def dtstartSeconds = date.getTime() / 1000  //epoch/1000
    def currentTimeSeconds = now() /1000  //epoch/1000
    
    return currentTimeSeconds - dtstartSeconds
}


//*************************
// parse(description)
// Called when the SmartThings Shield uses smartthing.send()
// 
//*************************
def parse(String description) {
  def value = zigbee.parse(description)?.text

    //log.debug "Raw: ${description}"
    //log.debug "Value: ${value}"

  if (value.indexOf("tick:") != -1 || value.indexOf("tock:") != -1) {
    oadrPoll() 
        if (S_pollInterval != value.substring(5).toInteger() ){
               log.debug "******************************"
               log.debug "OADR *** GUI Poll Interval (${S_pollInterval}) does not match Arduino Poll Interval (${value.substring(5).toInteger()})"
    }
        toggleButtonOn() 
  }
}


//*************************
// setPollInterval()
// DO NOT put code that goes out to the cloud after the zigbee.smartShield code - Message won't get to shield!
//*************************
def setPollInterval(){
  if (S_pollInterval != null){
      //log.debug "in setPollInterval"
        log.debug "******************************"
        toggleButtonOff()
    if (S_pollInterval < 10) {
            log.debug "OADR: *** Arduino poll interval set to to 10 seconds, requested value to low (${S_pollInterval})"
            zigbee.smartShield(text: "set:10").format()  //Don't allow less that 10
        }
        else {
            log.debug "OADR: *** Arduino poll interval set to to ${S_pollInterval} seconds"
            zigbee.smartShield(text: "set:${S_pollInterval}").format()
        }
    }
}

def toggleButtonOff(){
  sendEvent(name: "toggle", value: "press") 
}

def toggleButtonOn(){
  sendEvent(name: "toggle", value: "enabled") 
}
               